/* cpufreq.c - requires GKrellmM 2.0.0 or higher
 *
 * author: Christoph Winkelmann, <chw@tks6.net>
 *
 * date         version   comment
 * 07/11/2003   0.1       initial release
 * 12/12/2003   0.2       added slider to set cpu frequency
 * 29/12/2003   0.2.1     get cpu frequency from /proc/cpuinfo if
 *                        /proc/sys/cpu/ unavailable
 * 31/12/2003   0.3       added numerical display of cpu frequency
 *                        display slider only
 * 11/01/2004   0.3.1     always get cpu frequency from /proc/cpuinfo
 *                        get khz_max with khz
 *                        save khz_max in configuration
 *                        hand over absolute frequency to cpufreqset
 * 28/01/2004   0.3.3     support for /sys interface
 * 31/01/2004   0.4       make economic and efficient use of available
 *                        interfaces
 * 15/03/2004   0.5       update plugin once a second only
 *                        display cpu temperature optionally
 *                        display cpufreq governor optionally
 *                        toggle governor when clicking on governor name
 *                        optionally set userspace governor when moving slider
 *                        reserve more height for text and slider
 * 16/03/2004   0.5.1     scaling factor for temperature added
 *
 *   gcc -Wall -fPIC -Wall `pkg-config --cflags gtk+-2.0` -c cpufreq.c
 *   gcc -shared -Wl -o cpufreq.so cpufreq.o
 *   gkrellm -p cpufreq.so
 *  
 *  Adapted from the official plugin-demos
 */

#include <gkrellm2/gkrellm.h>

/* name in the configuration tree */
#define  CONFIG_NAME	"CPUfreq"

/* theme subdirectory for custom images for this plugin  and
   gkrellmrc style name for custom settings
*/
#define  STYLE_NAME	"cpufreq"

static int interface_proc;
static int interface_sys;

static double khz_max;
static double khz;

static double temperature;
static char temperature_file[256];
static char temperature_scale_text[256];
static double temperature_scale;
static int tempfile_offset;

static const int length = BUFSIZ;
static char governor[256];
static char empty[1];
static char* governor_text;

static GkrellmMonitor* monitor;
static GkrellmPanel* panel;
static GkrellmTicks* pGK;

static GkrellmDecal* text_decal;
static GkrellmDecal* text_decal_gov;

static GkrellmKrell* slider_krell;
static GkrellmKrell* slider_in_motion;
static double slider_value;

static gint temp_enable;
static gint gov_enable;
static gint gov_enable_current;
static gint slider_userspace_enable;
static gint style_id;

static void read_governor() {

  FILE* fp;
  int success = 0;
  if (interface_sys) {
    fp = fopen( "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "r" );
    if (fp) {
      char input[length];
      if (fgets(input, length, fp)) {
        strcpy(governor, input);
        governor[strlen(governor)-1] = 0;
        success = 1;
      }
      fclose( fp );
    }
  }
  if (!success) {
    fp = fopen( "/proc/cpufreq", "r" );
    if (fp) {
      char input[length];
      if (fgets(input, length, fp)) {
        if (fgets(input, length, fp)) {
          strcpy(governor, input+62);
          governor[strlen(governor)-1] = 0;
          success = 1;
        }
      }
      fclose( fp );
    }
  }
  if (!success) {
    strcpy(governor, "-");
  }
}

static void read_khz() {
  khz = 0;
  FILE* fp;

  /* current freq */

  if (interface_sys) {
    if (!gov_enable_current) {
      /* otherwise it has already been read in this update step */
      read_governor();
    }
    char suffix[length];
    if (strcmp(governor, "userspace") == 0) {
      strcpy(suffix, "setspeed");
    } else if (strcmp(governor, "performance") == 0) {
      strcpy(suffix, "max_freq");
    } else if (strcmp(governor, "powersave") == 0) {
      strcpy(suffix, "min_freq");
    } else {
      strcpy(suffix, "none");
    }
    if (strcmp(suffix, "none")) {
      char filename[length];
      sprintf(filename, "/sys/devices/system/cpu/cpu0/cpufreq/scaling_%s",
              suffix);
      fp = fopen(filename, "r");
      if (fp) {
        char input[length];
        if (fgets(input, length, fp)) {
          khz = atof(input);
        }
        fclose( fp );
      }
    }
  } 
  if (khz == 0) {
    if (interface_proc) {
      fp = fopen( "/proc/sys/cpu/0/speed", "r" );
      if (fp) {
        char input[length];
        if (fgets(input, length, fp)) {
          khz = atof(input);
        }
        fclose( fp );
      }
    }
    if (khz == 0) {
      fp = fopen( "/proc/cpuinfo", "r" );
      if (fp) {
        char input[length];
        while ( fgets( input, length, fp ) ) {
          if ( strncmp(input, "cpu MHz", 7)==0 ) {
            khz = 1000*atof(input+11);
          }
        }
        fclose( fp );
      }
    }
  }

  /* max freq */

  if (!interface_sys && interface_proc) {
    double khz_max_proc = 0;
    fp = fopen( "/proc/sys/cpu/0/speed-max", "r" );
    if ( fp ) {
      char input[length];
      if ( fgets( input, length, fp ) ) {
        khz_max_proc = atof(input);
      }
      fclose( fp );
    }
    khz_max = khz_max_proc > khz_max ? khz_max_proc : khz_max;
  }

  khz_max = khz > khz_max ? khz : khz_max;

}

static void read_temperature() {
  if (temp_enable) {
    temperature = 0;
    FILE* fp;

    fp = fopen( temperature_file, "r" );
    if (fp) {
      char input[length];
      char *inpnumber;
      char **inp;
      if ( fgets( input, length, fp ) ) {
        inpnumber = input + tempfile_offset;
        inp = &inpnumber;
        temperature = atof(strsep(inp, " ")) * temperature_scale;
      }
      fclose( fp );
    }
  }
}

static void update_plugin() {
  if(!pGK->second_tick) {
    return;
  }

  if (gov_enable_current) {
    read_governor();
  }
  read_khz();
  read_temperature();

  if (!slider_in_motion) {
    int krellpos = slider_krell->w_scale*khz/khz_max;
    gkrellm_update_krell(panel, slider_krell, krellpos);
  }
  
  char theText[length];
  if (temp_enable) {
    sprintf(theText, "%d, %1.3gC", (int)(khz/1000+0.5), temperature);
  } else {
    sprintf(theText, "%d MHz", (int)(khz/1000+0.5));
  }
  text_decal->x_off = 0;
  gkrellm_draw_decal_text(panel, text_decal, theText, -1);
  text_decal_gov->x_off = 0;
  gkrellm_draw_decal_text(panel, text_decal_gov, governor_text, -1);
  
  gkrellm_draw_panel_layers(panel);
}

static void update_slider_position(GkrellmKrell *k, gint x_ev) {
  gint	x;

  /* Consider only x_ev values that are inside the dynamic range of the
     krell.
  */
  x = x_ev - k->x0;
  if (x < 0)
  	x = 0;
  if (x > k->w_scale)
  	x = k->w_scale;
  gkrellm_update_krell(panel, k, x);
  gkrellm_draw_panel_layers(panel);

  slider_value = (double) x / (double) k->w_scale;  /* ranges 0.0 - 1.0 */
}

static gint cb_panel_motion(GtkWidget* widget, GdkEventMotion* ev,
                            gpointer data) {
  GdkModifierType state;
  
  if (slider_in_motion)	{
    /* Check if button is still pressed, in case missed button_release*/
  	state = ev->state;
  	if (!(state & GDK_BUTTON1_MASK))
  		slider_in_motion = NULL;
  	else
  		update_slider_position(slider_in_motion, (gint) ev->x);
  }
  return TRUE;
}

static gint cb_panel_release(GtkWidget* widget, GdkEventButton* ev,
                             gpointer data) {
  if (slider_in_motion) {
    char cmd[length];
    if (slider_userspace_enable) {
      sprintf(cmd, "sudo /usr/sbin/cpufreqsetgovernor userspace");
      system(cmd);
    }
    sprintf(cmd, "sudo /usr/sbin/cpufreqset %d", (int)(slider_value*khz_max));
    system(cmd);
  }
  slider_in_motion = NULL;
  return TRUE;
}

static gint cb_panel_press(GtkWidget* widget, GdkEventButton* ev,
                           gpointer data) {
  GkrellmKrell *k;

  if (ev->button == 3) {
  	gkrellm_open_config_window(monitor);
  	return TRUE;
  }
  /* Check if button was pressed within the space taken
     up by the slider krell.
  */
  slider_in_motion = NULL;
  k = slider_krell;
  if ( ev->x > k->x0 && ev->x <= k->x0 + k->w_scale
       && ev->y >= k->y0 && ev->y <= k->y0 + k->h_frame
       )
  	slider_in_motion = k;

  if (slider_in_motion)
  	update_slider_position(slider_in_motion, (gint) ev->x);

  /* Check if button was pressed within the space taken
     up by the governor text
  */
  if (gov_enable_current && 
      ev->x > text_decal_gov->x &&
      ev->x <= text_decal_gov->x + text_decal_gov->w &&
      ev->y >= text_decal_gov->y &&
      ev->y <= text_decal_gov->y + text_decal_gov->h) {
    /* toggle governor */
    read_governor();
    char cmd[length];
    char newgovernor[length];
    if (strcmp(governor, "userspace") == 0) {
      sprintf(newgovernor, "performance");
    } else if (strcmp(governor, "performance") == 0) {
      sprintf(newgovernor,"powersave");
    } else if (strcmp(governor, "powersave") == 0) {
      sprintf(newgovernor, "userspace");
    } else {
      sprintf(newgovernor, "performance");
    }
    sprintf(cmd, "sudo /usr/sbin/cpufreqsetgovernor %s", newgovernor);
    system(cmd);
  }

  return TRUE;
}

static void cb_check_tempfile_offset(GtkWidget *widget, GtkSpinButton *spin) {
  tempfile_offset = gtk_spin_button_get_value_as_int(spin);
}

static gint panel_expose_event(GtkWidget* widget, GdkEventExpose* ev) {
  gdk_draw_pixmap(widget->window,
                  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                  panel->pixmap, ev->area.x, ev->area.y, ev->area.x,
                  ev->area.y, ev->area.width, ev->area.height);
  return FALSE;
}

static void create_plugin(GtkWidget* vbox, gint first_create) {
  GkrellmStyle* style;
  GkrellmPiximage* krell_image;
  GkrellmTextstyle* ts;
  gint y;

  /* called when GKrellM is re-built at every theme or horizontal
     size change
  */
  if (first_create)
  	panel = gkrellm_panel_new0();

  style = gkrellm_meter_style(style_id);

  /* text */
  ts = gkrellm_meter_textstyle(style_id);

  /* the governor text */

  text_decal_gov = gkrellm_create_decal_text(panel,
                                             "abcdefghijklmnopqrstuvwxyz", ts,
                                             style,
     -1,     /* x = -1 places at left margin */
     -1,     /* y = -1 places at top margin */
     -1);    /* w = -1 makes decal the panel width minus margins */
  if (gov_enable) {
    governor_text = governor;
    y = text_decal_gov->y + text_decal_gov->h + 1;
  } else {
    governor_text = empty;
    y = -1;
  }
  gov_enable_current = gov_enable;

  /* the frequency text */
  text_decal = gkrellm_create_decal_text(panel, "0123456789  MHz", ts, style,
      -1,     /* x = -1 places at left margin */
       y,     /* y = -1 places at top margin */
      -1);    /* w = -1 makes decal the panel width minus margins */

  y = text_decal->y + text_decal->h + 1;

  /* the slider */
  krell_image = gkrellm_krell_slider_piximage();
  gkrellm_set_style_slider_values_default(style,
  	y,     /* y offset */
  	0, 0); /* Left and right krell margins */
  slider_krell = gkrellm_create_krell(panel, krell_image, style);
  y = slider_krell->y0 + slider_krell->h_frame + 2;

  gkrellm_monotonic_krell_values(slider_krell, FALSE);
  gkrellm_set_krell_full_scale(slider_krell, slider_krell->w_scale, 1);
  gkrellm_update_krell(panel, slider_krell, 0);

  read_khz();
  update_slider_position(slider_krell, slider_krell->w_scale*khz/khz_max);

  /* configure and create the panel  */
  gkrellm_panel_configure(panel, "", style);
  gkrellm_panel_create(vbox, monitor, panel);

  if (first_create) {
    g_signal_connect(G_OBJECT (panel->drawing_area), "expose_event",
                     G_CALLBACK(panel_expose_event), NULL);
    g_signal_connect(G_OBJECT(panel->drawing_area), "button_press_event",
                     G_CALLBACK(cb_panel_press), NULL );
    g_signal_connect(G_OBJECT(panel->drawing_area), "button_release_event",
                     G_CALLBACK(cb_panel_release), NULL );
    g_signal_connect(G_OBJECT(panel->drawing_area), "motion_notify_event",
                     G_CALLBACK(cb_panel_motion), NULL);
  }
}

/* Configuration */

#define PLUGIN_CONFIG_KEYWORD  STYLE_NAME

static GtkWidget* temp_enable_button;
static GtkWidget* gov_enable_button;
static GtkWidget* slider_userspace_enable_button;
static GtkWidget* temperature_file_entry;
static GtkWidget* temperature_scale_entry;

static void save_plugin_config(FILE *f) {
  fprintf(f, "%s khz_max %d\n", PLUGIN_CONFIG_KEYWORD, (int)khz_max);
  fprintf(f, "%s temp_enable %d\n", PLUGIN_CONFIG_KEYWORD, temp_enable);
  fprintf(f, "%s gov_enable %d\n", PLUGIN_CONFIG_KEYWORD, gov_enable);
  fprintf(f, "%s slider_userspace_enable %d\n", PLUGIN_CONFIG_KEYWORD,
          slider_userspace_enable);
  fprintf(f, "%s temperature_file %s\n", PLUGIN_CONFIG_KEYWORD,
          temperature_file);
  fprintf(f, "%s tempfile_offset %d\n", PLUGIN_CONFIG_KEYWORD,
          tempfile_offset);
  fprintf(f, "%s temperature_scale %g\n", PLUGIN_CONFIG_KEYWORD,
          temperature_scale);
}

static void load_plugin_config(gchar *arg) {
  gchar	config[64], item[256];
  gint	n;
  int khz_max_config;

  n = sscanf(arg, "%s %[^\n]", config, item);
  if (n == 2) {
    if (strcmp(config, "khz_max") == 0) {
      sscanf(item, "%d\n", &khz_max_config);
      if (khz_max_config > khz_max) {
        khz_max = khz_max_config;
      }
    } else if (strcmp(config, "temp_enable") == 0) {
      sscanf(item, "%d", &temp_enable);
    } else if (strcmp(config, "slider_userspace_enable") == 0) {
      sscanf(item, "%d", &slider_userspace_enable);
    } else if (strcmp(config, "gov_enable") == 0) {
      sscanf(item, "%d", &gov_enable);
      gov_enable_current = gov_enable;
      governor_text = gov_enable ? governor : empty;
    } else if (strcmp(config, "temperature_file") == 0) {
      strcpy(temperature_file, item);
    } else if (strcmp(config, "tempfile_offset") == 0) {
      sscanf(item, "%d", &tempfile_offset);
    } else if (strcmp(config, "temperature_scale") == 0) {
      sscanf(item, "%lg", &temperature_scale);
    }
  }
}

static void apply_plugin_config(void) {
  temp_enable = GTK_TOGGLE_BUTTON(temp_enable_button)->active;
  gov_enable = GTK_TOGGLE_BUTTON(gov_enable_button)->active;
  slider_userspace_enable =
    GTK_TOGGLE_BUTTON(slider_userspace_enable_button)->active;
  strcpy(temperature_file,
         gkrellm_gtk_entry_get_text(&temperature_file_entry));
  strcpy(temperature_scale_text,
         gkrellm_gtk_entry_get_text(&temperature_scale_entry));
  sscanf(temperature_scale_text, "%lg", &temperature_scale);
}

static gchar  *plugin_info_text[] = {
  "<h>CPU frequency plugin\n",
  "Author: Christoph Winkelmann, ",
  "<ul> chw@tks6.net\n\n",
  "Enabling and disabling governor display only takes effect after change\n",
  "of theme, change of width, or restart.\n\n",
  "The maximal frequency found during this run is saved for the next run\n",
  "when this page is viewed.\n\n",
  "Any suggestions are welcome!",
  /* "\n\tPut any documentation here to explain your plugin.\n",
     "\tText can be ",
     "<b>bold",
     " or ",
     "<i>italic",
     " or ",
     "<ul>underlined.\n",
  */
};

static void create_plugin_tab(GtkWidget *tab_vbox) {
  GtkWidget* tabs;
  GtkWidget* vbox;
  GtkWidget* vbox1;
  GtkWidget* text;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* label;
  gint i;

  tabs = gtk_notebook_new();
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
  gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);

  /* -- Options tab -- */
  vbox = gkrellm_gtk_framed_notebook_page(tabs, "Options");

  /* -- Governor options -- */
  vbox1 = gkrellm_gtk_framed_vbox(vbox, "CPUfreq governor", 4, FALSE, 0, 2);
  gkrellm_gtk_check_button(vbox1, &gov_enable_button, gov_enable,
                           FALSE, 0, "Show governor (see Info tab)");
  gkrellm_gtk_check_button(vbox1, &slider_userspace_enable_button,
                           slider_userspace_enable, FALSE, 0,
                           "Set userspace governor when moving slider");

  vbox1 = gkrellm_gtk_framed_vbox(vbox, "CPU temperature", 4, FALSE, 0, 2);
  gkrellm_gtk_check_button(vbox1, &temp_enable_button, temp_enable, FALSE, 0,
                           "Show CPU temperature");

  /* -- Temperature options -- */
  table = gtk_table_new(5, 2, FALSE /*non-homogeneous*/);
  gtk_table_set_col_spacings(GTK_TABLE(table), 2);
  gtk_box_pack_start(GTK_BOX(vbox1), table, FALSE, FALSE, 2);

  /* -- Temperature: Filename -- */
  hbox = gtk_hbox_new(TRUE, 0);
  /* Attach left right top bottom */
  gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 0, 1,
                   GTK_SHRINK, GTK_SHRINK, 0, 0);
  label = gtk_label_new(_("from file"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
  temperature_file_entry = gtk_entry_new();
  gtk_entry_set_max_length(GTK_ENTRY(temperature_file_entry), 255);
  gtk_table_attach_defaults(GTK_TABLE(table), temperature_file_entry,
                            1, 2, 0, 1);
  gtk_entry_set_text(GTK_ENTRY(temperature_file_entry), temperature_file);

  /* -- Temperature: Offset -- */
  hbox = gtk_hbox_new(TRUE, 0);
  /* Attach left right top bottom */
  gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 1, 2,
                   GTK_SHRINK, GTK_SHRINK, 0, 0);
  label = gtk_label_new(_("file offset"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 2, 1, 2);
  gkrellm_gtk_spin_button(hbox, NULL, tempfile_offset/*value*/, 0./*low*/,
                          65535./*high*/, 1/*step0*/, 10/*step1*/,
                          0 /*digits*/, 0 /*width*/,
                          cb_check_tempfile_offset /* (*cb_func)()*/,
                          NULL /*data*/,
                          FALSE /*right_align*/, "characters");

  /* -- Temperature: Scaling factor -- */
  hbox = gtk_hbox_new(TRUE, 0);
  /* Attach left right top bottom */
  gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 2, 3,
                   GTK_SHRINK, GTK_SHRINK, 0, 0);
  label = gtk_label_new(_("scale factor"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 4);
  temperature_scale_entry = gtk_entry_new();
  gtk_entry_set_max_length(GTK_ENTRY(temperature_scale_entry), 255);
  gtk_table_attach_defaults(GTK_TABLE(table), temperature_scale_entry,
                            1, 2, 2, 3);
  sprintf(temperature_scale_text, "%g", temperature_scale);
  gtk_entry_set_text(GTK_ENTRY(temperature_scale_entry),
                     temperature_scale_text);

  /* -- Info tab -- */
  vbox = gkrellm_gtk_framed_notebook_page(tabs, "Info");
  text = gkrellm_gtk_scrolled_text_view(vbox, NULL, GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC);
  for (i = 0; i < sizeof(plugin_info_text)/sizeof(gchar *); ++i)
  	gkrellm_gtk_text_view_append(text, plugin_info_text[i]);
}

/* The monitor structure tells GKrellM how to call the plugin routines.
*/
static GkrellmMonitor  plugin_mon = {
  CONFIG_NAME,        /* Title for config clist.   */
  0,                  /* Id,  0 if a plugin       */
  create_plugin,      /* The create function      */
  update_plugin,      /* The update function      */
  create_plugin_tab,  /* The config tab create function   */
  apply_plugin_config,/* Apply the config function        */

  save_plugin_config, /* Save user config			*/
  load_plugin_config, /* Load user config			*/
  PLUGIN_CONFIG_KEYWORD, /* config keyword			*/

  NULL,               /* Undefined 2	*/
  NULL,               /* Undefined 1	*/
  NULL,               /* private	*/

  MON_CPU,            /* Insert plugin before this monitor			*/

  NULL,               /* Handle if a plugin, filled in by GKrellM     */
  NULL                /* path if a plugin, filled in by GKrellM       */
};


/* All GKrellM plugins must have one global routine named init_plugin()
   which returns a pointer to a filled in monitor structure.
*/

GkrellmMonitor* gkrellm_init_plugin(void)

{
  /* If this next call is made, the background and krell images for this
     plugin can be custom themed by putting bg_meter.png or krell.png in the
     subdirectory STYLE_NAME of the theme directory.  Text colors (and
     other things) can also be specified for the plugin with gkrellmrc
     lines like:  StyleMeter  STYLE_NAME.textcolor orange black shadow
     If no custom theming has been done, then all above calls using
     style_id will be equivalent to style_id = DEFAULT_STYLE_ID.
  */
  pGK = gkrellm_ticks();
  style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);
  monitor = &plugin_mon;

  /* determine interface availability */
  FILE* fp;

  interface_proc = 0;
  fp = fopen( "/proc/sys/cpu/0/speed", "r" );
  if (fp) {
    interface_proc = 1;
    fclose(fp);
  }

  interface_sys = 0;
  fp = fopen( "/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq", "r" );
  if (fp) {
    interface_sys = 1;
    char input[length];
    if (fgets(input, length, fp)) {
      khz_max = atof(input);
    }
    fclose(fp);
  }

  read_khz();
  read_temperature();

  strcpy(empty, "");
  if (gov_enable_current) {
    read_governor();
  } else {
    strcpy(governor, "");
  }
  return &plugin_mon;
}
